%{
#include <stdio.h>
#include <string.h>
 
extern int yyparse( void );
extern int yylex (void);
extern int yylineno;


static int expr_result = 0;
static int yyerrors   = 0;


void yyerror(const char *str)
{
        fprintf(stderr,"line %d: error: %s\n", yylineno, str);
	yyerrors++;
}
 
int yywrap( void )
{
        return 1;
} 

int main( int argc, char **argv )
{
        if( yyparse() == 0 && yyerrors == 0 )
	{
		printf( "parsing succeeded\n" );
		printf( "result: %d\n", expr_result );
	} else
	{
		fprintf( stderr, "parsing failed\n" );
	}
	return 0;
} 
%}

%union {
	int n;
}

%token PLUS MINUS MUL DIV MOD OPEN CLOSE TOKERR
%token <n> NUMBER

%type <n> expr term factor

%start oneexpr

%%
oneexpr      :  expr            { expr_result = $1; }
             ;

expr         : expr PLUS term   { $$ = $1 + $3; }
             | expr MINUS term  { $$ = $1 - $3; }
             | term
             ;

term         : term MUL factor  { $$ = $1 * $3; }
             | term DIV factor  { $$ = $1 / $3; }
             | term MOD factor  { $$ = $1 % $3; }
             | factor
             ;

factor       : NUMBER
             | OPEN expr CLOSE  { $$ = $2; }
             ;
